/**
 * Copyright Copyright 2008-14 Simon Andrews
 *
 *    This file is part of FocalPoint.
 *
 *    FocalPoint is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    FocalPoint is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.FocalPoint.GUI.Thumbnail.Resizer;

import ij.ImagePlus;
import ij.ImageStack;
import ij.process.StackProcessor;

import java.util.Enumeration;
import java.util.Vector;

import uk.ac.bbsrc.babraham.FocalPoint.GUI.Thumbnail.ThumbnailPanel;

public class ImageResizer implements Runnable{

	/*
	 * This class acts as a common resource to do the resizing of
	 * images.  It means the rest of the interface can be getting
	 * on with other stuff whilst the images are waiting to be
	 * resized.
	 * 
	 * It's implemented with a background thread which processes
	 * requests one at a time.
	 */
	
	private static final ImageResizer resizer = new ImageResizer();
	private Vector<ResizeInformation> queue = new Vector<ResizeInformation>();
	private Vector<ResizeQueueListener> listeners = new Vector<ResizeQueueListener>();
	private boolean isIdle = true;
	private StackProcessor sp;
	private ImageStack resizedStack;
	
	private ImageResizer () {
		Thread t = new Thread(this);
		t.start();
	}
	
	public void addListener (ResizeQueueListener l) {
		if (l != null && ! listeners.contains(l)) listeners.add(l);
	}

	public void removeListener (ResizeQueueListener l) {
		if (l != null && listeners.contains(l)) listeners.remove(l);
	}

	
	public static ImageResizer getInstance () {
		return resizer;
	}

	public void resizeImages (ImagePlus [] ip, ResizeListener rl) {
		
		/*
		 * If there's already a job in the queue for this resize listener
		 * then we can delete it and substitute this one instead
		 */
		
		Enumeration<ResizeInformation> e = queue.elements();
		
		while (e.hasMoreElements()) {
			ResizeInformation r = e.nextElement();
			if (r.rl == rl) {

				queue.remove(r);

				// Close the ImagePlus objects for the object we're about
				// to discard.
				for (int i=0;i<r.ip.length;i++) {
					r.ip[i].close();
				}
			}
		}
		
		
		queue.add(new ResizeInformation(ip,rl));
	}
	
	private void processRequest (ResizeInformation r) {
		
		ImagePlus [] resizedIPs = new ImagePlus[r.ip.length];
		
		for (int c=0;c<r.ip.length;c++) {

			try {
				Thread.sleep(5);
			} 
			catch (InterruptedException e) {}
			
			sp = new StackProcessor(r.ip[c].getStack(),r.ip[c].getProcessor());

			resizedStack = sp.resize(ThumbnailPanel.thumbnailWidth, (int)(r.ip[c].getHeight()*((double)ThumbnailPanel.thumbnailWidth/r.ip[c].getWidth())));
			
			resizedIPs[c] = new ImagePlus(r.ip[c].getTitle(),resizedStack);
			r.ip[c].killStack();
//			r.ip[c].killProcessor();
			r.ip[c].close();
			
		}

		r.rl.ImageResized(resizedIPs);

		/* 
		 * Just trying to eliminate any possible refs causing a memory leak
		 */
		r.ip = null;
		r.rl = null;
	}
	
	
	public void run() {

		while (true) {
			// Check if there's a job waiting
			if (queue.size() > 0) {
				// Check to see if we've just become active
				if (isIdle) {
					isIdle = false;
					Enumeration<ResizeQueueListener> e = listeners.elements();
					while (e.hasMoreElements()) {
						e.nextElement().isResizing();
					}
				}
				ResizeInformation r = (ResizeInformation)queue.elementAt(0);
				queue.remove(r);
				processRequest(r);
			}
			else {
				// Check to see if we've just gone idle
				if (!isIdle) {
					isIdle = true;
					Enumeration<ResizeQueueListener> e = listeners.elements();
					while (e.hasMoreElements()) {
						e.nextElement().isIdle();
					}
				}
			}
			
			try {
				Thread.sleep(50);
			} 
			catch (InterruptedException e) {}
		}
		
	}
	
	private class ResizeInformation {
		ImagePlus [] ip;
		ResizeListener rl;
		
		public ResizeInformation (ImagePlus [] ip, ResizeListener rl) {
			this.ip = ip;
			this.rl = rl;
		}
	}
	
}
